﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Security.Cryptography;

using System.Threading;

namespace Biclustering
{
    public partial class frmBiclustering : Form
    {
        List<Cluster> mejoresClusters;
        List<string> Genes;
        List<string> Condiciones;

        double[,] datos;
        int nCondiciones=0;
        int nGenes=0;
        int minimoGenes = 0;
        int maximoGenes = 0;
        int minimoCondiciones = 0;
        int maximoCondiciones = 0;
        int nGeneraciones = 0;
        int nEjecuciones = 0;
        double calidadMinima = 0;      

        int nIndividuos = 100;
        int pSeleccion = 90;
        int pCruza = 100;
        int pMutacion = 50;
        string nombreArchivo = "";
        string archivoGenes = "";
        string archivoCondiciones = "";

        RNGCryptoServiceProvider rng = new RNGCryptoServiceProvider();
        byte[] datoRand = new byte[4];


        public frmBiclustering()
        {
            InitializeComponent();
        }
       
        private void cargarDatos()
        {
            datos = new double[nGenes, nCondiciones];
            StreamReader archivo = new StreamReader(txtDatos.Text.ToString().Trim());
            char car;
            string dato = "";
            int Gen = 0;
            int Condicion = 0;
            while (archivo.EndOfStream == false)
            {
                car = (char)archivo.Read();
                if (car == '-' || car == ' ' || car == '\n' || car == '\t')
                {
                    if (dato.Length > 0)
                    {
                        datos[Gen, Condicion] = Convert.ToDouble(dato.ToString());
                        Condicion++;
                    }
                    dato = "";
                    if (car == '\n')
                    {
                        Gen++;
                        Condicion = 0;
                    }
                    if (car == '-')
                    {
                        dato += car.ToString();
                    }
                }
                else
                {
                    dato += car.ToString();
                }
            }
            archivo.Close();

            Genes = new List<string>();
            StreamReader entradaGenes = new StreamReader(archivoGenes.ToString().Trim());
            string gen = "";
            while (entradaGenes.EndOfStream == false)
            {
                car = (char)entradaGenes.Read();
                if (car == ' ' || car == '\n' || car == '\t')
                {
                    if (gen.Length > 0)
                    {
                        Genes.Add(gen.ToString());
                    }
                    gen = "";
                }
                else
                {
                    gen += car.ToString();
                }
            }
            if (gen.Length > 0)
            {
                Genes.Add(gen.ToString());
            }
            entradaGenes.Close();

            Condiciones = new List<string>();
            StreamReader entradaCondiciones = new StreamReader(archivoCondiciones.ToString().Trim());

            string condicion = "";
            while (entradaCondiciones.EndOfStream == false)
            {
                car = (char)entradaCondiciones.Read();
                if (car == ' ' || car == '\n' || car == '\t')
                {
                    if (condicion.Length > 0)
                    {
                        Condiciones.Add(condicion.ToString());
                    }
                    condicion = "";
                }
                else
                {
                    condicion += car.ToString();
                }
            }
            entradaCondiciones.Close();

        }
        
        private void cmdDatos_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlgArchivo = new OpenFileDialog();
            dlgArchivo.ShowDialog();
            txtDatos.Text = dlgArchivo.FileName;
        }

        private void salidaInicial()
        {
            StreamWriter salida = new StreamWriter(nombreArchivo + ".txt", true);
            string cadena = txtDatos.Text;
            cadena = cadena.Substring(cadena.LastIndexOf('\\') + 1, cadena.Length - (cadena.LastIndexOf('\\') + 1));
            salida.WriteLine();
            salida.WriteLine("-----------------------------------------------------------------------------------");
            salida.WriteLine("Data: " + cadena);
            salida.WriteLine("Genes: " + nGenes.ToString());
            salida.WriteLine("Conditions: " + nCondiciones.ToString());
            salida.WriteLine("Individuals: " + nIndividuos.ToString());
            salida.WriteLine("Generations without improvement: " + nGeneraciones.ToString());
            salida.WriteLine("Selection: " + pSeleccion.ToString() + "%");
            salida.WriteLine("Crossover: " + pCruza.ToString() + "%");
            salida.WriteLine("Mutation: " + pMutacion.ToString() + "%");
            salida.Close();
        }

        private void salidaInicialGenético(int nE, int numGenes, int numCondiciones, string horaInicio)
        {
            StreamWriter salida = new StreamWriter(nombreArchivo + ".txt", true);
            salida.WriteLine("-----------------------------------------------------------------------------------");
            salida.WriteLine("Run No. " + (nE + 1).ToString());
            salida.WriteLine("Start: " + horaInicio);
            salida.Close();
        }

        private void cmdEjecutar_Click(object sender, EventArgs e)
        {

            nGenes = Convert.ToInt32(txtGenes.Text.ToString());
            nCondiciones = Convert.ToInt32(txtCondiciones.Text.ToString());
            calidadMinima = 1-(Convert.ToSingle(txtCalidadBiclusters.Text.ToString())/100);      

            nEjecuciones = Convert.ToInt32(txtEjecuciones.Text.ToString());
            nGeneraciones = Convert.ToInt32(txtGeneracionesSinMejora.Text.ToString());

            minimoGenes = Convert.ToInt32(txtMinimoGenes.Text.ToString());
            maximoGenes = Convert.ToInt32(txtMaximoGenes.Text.ToString());
            minimoCondiciones = Convert.ToInt32(txtMinimoCondiciones.Text.ToString());
            maximoCondiciones = Convert.ToInt32(txtMaximoCondiciones.Text.ToString());
            
            nombreArchivo = txtSalida.Text.ToString().Trim();
            nombreArchivo = nombreArchivo.Replace(".txt", "");

            archivoGenes=txtArchivoGenes.Text.ToString();
            archivoCondiciones=txtArchivoCondiciones.Text.ToString();

            mejoresClusters = new List<Cluster>();
            Genes = new List<string>();
            Condiciones = new List<string>();

            salidaInicial();
            cargarDatos();

            for (int nE = 0; nE < nEjecuciones; nE++)
            {

                lblRun.Text = "Run # " + Convert.ToString(nE+1);
                lblRun.Refresh();

                rng.GetNonZeroBytes(datoRand);
                int valor = BitConverter.ToInt32(datoRand, 0);
                int umbralGenes = minimoGenes+ Math.Abs(valor % (maximoGenes-minimoGenes+1));
                int umbralCondiciones = minimoCondiciones+ Math.Abs(valor % (maximoCondiciones-minimoCondiciones+1));
         
                Poblacion poblacion = new Poblacion(datos, nIndividuos, umbralGenes, umbralCondiciones);
                poblacion.calculaAptitud();

                lstResultados.Items.Clear();
                lstResultados.Items.Add("       Total            Without       Number     Number  of      Size         VF    ");
                lstResultados.Items.Add("  Generations   Improvement   of Genes   Conditions   Bicluster   Function ");
                lstResultados.Items.Add("--------------------------------------------------------------------------------------");

                int nG = 0;
                int cont = 0;
                double mejorVariación = -1;
                string horaInicio = DateTime.Now.ToString();

                while (cont < nGeneraciones)
                {
                    nG++;

                    poblacion.selección(pSeleccion);
                    poblacion.cruza(pCruza);                             
                    poblacion.mutacion(pMutacion);                       
                    poblacion.calculaAptitud();

                    if (mejorVariación == -1) mejorVariación = poblacion.clusters[0].variación;
                    if (poblacion.clusters[0].variación < mejorVariación) { mejorVariación = poblacion.clusters[0].variación; cont = 0; } else cont++;

                    desplegarEstadísticaCluster(nG, cont, poblacion.clusters[0]);

                    if (cont == nGeneraciones && mejorVariación > (calidadMinima * (poblacion.clusters[0].numCondiciones - 1)* (2 * poblacion.clusters[0].numGenes - 2)))
                    {
                        rng.GetNonZeroBytes(datoRand);
                        valor = BitConverter.ToInt32(datoRand, 0);
                        umbralGenes = minimoGenes + Math.Abs(valor % (maximoGenes - minimoGenes + 1));
                        umbralCondiciones = minimoCondiciones + Math.Abs(valor % (maximoCondiciones - minimoCondiciones + 1));

                        poblacion = new Poblacion(datos, nIndividuos, umbralGenes, umbralCondiciones);
                        poblacion.calculaAptitud();

                        nG = 0;
                        cont = 0;
                        mejorVariación = -1;
                        horaInicio = DateTime.Now.ToString();

                    }
                }

                salidaInicialGenético(nE, umbralGenes, umbralCondiciones,horaInicio);
                desplegarEstadísticaGenético(nombreArchivo, archivoGenes, archivoCondiciones, poblacion.clusters[0], nG, nE);
                
                mejoresClusters.Add(poblacion.clusters[0].copia());
            }

            StreamWriter salida = new StreamWriter(nombreArchivo+".txt", true);
            salida.WriteLine();
            salida.WriteLine("\n-----------------------------------------------------------------------------------");
            salida.WriteLine("Biclusters with no more than 25% overlap:");
                        
            for (int i = 0; i < mejoresClusters.Count - 1; i++)
            {
                for (int j = i + 1; j < mejoresClusters.Count; j++)
                {
                    if (mejoresClusters[i] != null && mejoresClusters[j] != null)
                    {
                        int cgi = 0;
                        int cgj = 0;
                        int cge = 0;
                        for (int g = 0; g < nGenes; g++)
                        {
                            if (mejoresClusters[i].genesCondiciones[g] == 1) cgi++;
                            if (mejoresClusters[j].genesCondiciones[g] == 1) cgj++;
                            if (mejoresClusters[i].genesCondiciones[g] == 1 && mejoresClusters[j].genesCondiciones[g] == 1) cge++;
                        }
                        if (cge > cgi * 0.25 || cge > cgj * 0.25)
                        {
                            if (cgi >= cgj)
                            {
                                mejoresClusters[j] = null;
                            }
                            else
                            {
                                mejoresClusters[i] = null;
                                break;
                            }
                        }
                    }
                }
            }
            
            bool sw = false;
            for (int i = 0; i < mejoresClusters.Count; i++)
            {
                if (mejoresClusters[i] != null)
                {
                    if (sw == true) salida.Write(","); else sw = true;
                    salida.Write((i+1).ToString().Trim());
                }
            }
            salida.Close();
            
        }

        private void desplegarEstadísticaCluster(int nG, int nGSM, Cluster cluster)
        {
            string cadena = String.Format("  {0,11}   ", nG);
            cadena += String.Format("     {0,11}   ", nGSM);
            cadena += String.Format("     {0,8:F0} ", cluster.numGenes);
            cadena += String.Format("     {0,10:F0}  ", cluster.numCondiciones);
            cadena += String.Format("   {0,11:F0} ", cluster.numGenes*cluster.numCondiciones);
            cadena += String.Format(" {0,10:F3} ", cluster.calculoFuncionVariacion());

            if (lstResultados.Items.Count>=23) lstResultados.Items.RemoveAt(22);
            lstResultados.Items.Insert(3, cadena);
            lstResultados.Update();
        }


        public void desplegarEstadísticaGenético(string archivo, string archivoGenes, string archivoCondiciones, Cluster mejorCluster, int nG, int nE)
        {
         
            StreamWriter salida = new StreamWriter(archivo.ToString().Trim() + ".txt", true);

            salida.WriteLine("Finish: " + DateTime.Now);
            salida.WriteLine("Generations Realized: " + nG.ToString());
            salida.WriteLine();
            salida.WriteLine("Best Bicluster:");
            salida.WriteLine();
            salida.WriteLine("  Genes Conditions    Size       VF  ");
            salida.WriteLine("----------------------------------------");

            string cadena = "";
            cadena += String.Format(" {0,5:F0} ", mejorCluster.numGenes);
            cadena += String.Format("   {0,4:F0}   ", mejorCluster.numCondiciones);
            cadena += String.Format("   {0,6:F0}  ", mejorCluster.numGenes * mejorCluster.numCondiciones);
            cadena += String.Format(" {0,8:F3} ", mejorCluster.variación);
            salida.WriteLine(cadena);

            salida.WriteLine();  
            salida.WriteLine("Bicluster:");
            salida.WriteLine();  
            for (int i = 0; i < nGenes + nCondiciones; i++)
            {
                salida.Write(mejorCluster.genesCondiciones[i].ToString());
            }
            
            salida.WriteLine();
            salida.WriteLine();
            salida.WriteLine("Genes: " + mejorCluster.numGenes.ToString().Trim());
            salida.Write("ID: ");
            for (int g = 0; g < nGenes; g++)
            {
                if (mejorCluster.genesCondiciones[g] == 1)
                {
                    salida.Write(Genes[g].Trim() + " ");
                }
            }
            salida.WriteLine();
            salida.WriteLine();
            salida.WriteLine("Conditions: " + mejorCluster.numCondiciones.ToString().Trim());
            salida.Write("ID: ");
            for (int c = 0; c < nCondiciones; c++)
            {
                if (mejorCluster.genesCondiciones[nGenes + c] == 1)
                {
                    salida.Write(Condiciones[c].Trim() + " ");
                }
            }
            salida.WriteLine();

            salida.Close();
        }

        private void cmdSalida_Click(object sender, EventArgs e)
        {
            SaveFileDialog dlgArchivo = new SaveFileDialog();
            dlgArchivo.ShowDialog();
            txtSalida.Text = dlgArchivo.FileName;
        }

        private void cmdGenes_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlgArchivo = new OpenFileDialog();
            dlgArchivo.ShowDialog();
            txtArchivoGenes.Text = dlgArchivo.FileName;
        }

        private void cmdCondiciones_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlgArchivo = new OpenFileDialog();
            dlgArchivo.ShowDialog();
            txtArchivoCondiciones.Text = dlgArchivo.FileName;
        }

    }
}
