﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Security.Cryptography;

namespace Biclustering
{
    public class Cluster
    {
        double[,] datos;
        int nGenesTotal = 0;
        int nCondicionesTotal = 0;
        public int[] genesCondiciones;
        public double variación;
        public int numGenes;
        public int numCondiciones;
 
        RNGCryptoServiceProvider rng = new RNGCryptoServiceProvider();
        byte[] dato = new byte[4];

        public Cluster(double[,] datos)
        {
            this.datos = datos;
            nGenesTotal = datos.GetLength(0);
            nCondicionesTotal = datos.GetLength(1);
            genesCondiciones = new int[nGenesTotal + nCondicionesTotal];
        }

        public void inicializa(int numGenes, int numCondiciones)
        {
            int valor;
            int condicion;
            int gen;
            for (int i = 0; i < nGenesTotal + nCondicionesTotal; i++)
            {
                genesCondiciones[i] = 0;
            }
            for (int c=0; c<numCondiciones; c++) {
                rng.GetNonZeroBytes(dato);
                valor = BitConverter.ToInt32(dato, 0);
                condicion = Math.Abs(valor % nCondicionesTotal);
                if (genesCondiciones[nGenesTotal + condicion] == 0)
                    genesCondiciones[nGenesTotal + condicion] = 1;
                else
                    c = c - 1;
            }
            numCondiciones=calculaNumCondiciones();
            for (int g = 0; g < numGenes; g++)
            {
                rng.GetNonZeroBytes(dato);
                valor = BitConverter.ToInt32(dato, 0);
                gen = Math.Abs(valor % nGenesTotal);
                if (genesCondiciones[gen] == 0)
                    genesCondiciones[gen] = 1;
                else
                    g = g - 1;
            }
            numGenes = calculaNumGenes();
        }

        public Cluster copia()
        {
            Cluster copia = new Cluster(datos);
            for (int i = 0; i < nGenesTotal + nCondicionesTotal; i++)
            {
                copia.genesCondiciones[i] = genesCondiciones[i];
            }
            copia.numGenes = numGenes;
            copia.numCondiciones = numCondiciones;
            copia.variación = variación;
            return copia;
        }

        public void mutacion()
        {
            int posMutacion;

            rng.GetNonZeroBytes(dato);
            int valor = BitConverter.ToInt32(dato, 0);
            int opción = Math.Abs(valor % 2);

            if (opción == 0)
            {
                rng.GetNonZeroBytes(dato);
                valor = BitConverter.ToInt32(dato, 0);
                posMutacion = Math.Abs(valor % nGenesTotal);

                if (genesCondiciones[posMutacion] == 1)
                {
                    genesCondiciones[posMutacion] = 0;
                }
                else
                {
                    genesCondiciones[posMutacion] = 1;
                }
            }
            else
            {
                rng.GetNonZeroBytes(dato);
                valor = BitConverter.ToInt32(dato, 0);
                posMutacion = Math.Abs(valor % nCondicionesTotal);

                if (genesCondiciones[nGenesTotal + posMutacion] == 1)
                {
                    genesCondiciones[nGenesTotal + posMutacion] = 0;
                }
                else
                {
                    genesCondiciones[nGenesTotal + posMutacion] = 1;
                }
            }
            numGenes = calculaNumGenes();
            numCondiciones = calculaNumCondiciones();
        }

        public Cluster cruza(Cluster otroCluster)
        {
            Cluster nuevoCluster = new Cluster(datos);

            int puntoCruzaGenes;
            int puntoCruzaCondiciones;
            int swGenes;
            int swCondiciones;

            rng.GetNonZeroBytes(dato);
            int valor = BitConverter.ToInt32(dato, 0);
            puntoCruzaGenes = Math.Abs(valor % nGenesTotal);

            rng.GetNonZeroBytes(dato);
            valor = BitConverter.ToInt32(dato, 0);
            swGenes = Math.Abs(valor % 2);

            rng.GetNonZeroBytes(dato);
            valor = BitConverter.ToInt32(dato, 0);
            puntoCruzaCondiciones = Math.Abs(valor % nCondicionesTotal);

            rng.GetNonZeroBytes(dato);
            valor = BitConverter.ToInt32(dato, 0);
            swCondiciones = Math.Abs(valor % 2);

            if (swGenes == 0)
            {
                for (int i = 0; i < puntoCruzaGenes; i++)
                {
                    nuevoCluster.genesCondiciones[i] = genesCondiciones[i];
                }
                for (int i = puntoCruzaGenes; i < nGenesTotal; i++)
                {
                    nuevoCluster.genesCondiciones[i] = otroCluster.genesCondiciones[i];
                }
            }
            else
            {
                for (int i = 0; i < puntoCruzaGenes; i++)
                {
                    nuevoCluster.genesCondiciones[i] = otroCluster.genesCondiciones[i];
                }
                for (int i = puntoCruzaGenes; i < nGenesTotal; i++)
                {
                    nuevoCluster.genesCondiciones[i] = genesCondiciones[i];
                }
            }

            if (swCondiciones == 0)
            {
                for (int i = 0; i < puntoCruzaCondiciones; i++)
                {
                    nuevoCluster.genesCondiciones[nGenesTotal + i] = genesCondiciones[nGenesTotal + i];
                }
                for (int i = puntoCruzaCondiciones; i < nCondicionesTotal; i++)
                {
                    nuevoCluster.genesCondiciones[nGenesTotal + i] = otroCluster.genesCondiciones[nGenesTotal + i];
                }
            }
            else
            {
                for (int i = 0; i < puntoCruzaCondiciones; i++)
                {
                    nuevoCluster.genesCondiciones[nGenesTotal + i] = otroCluster.genesCondiciones[nGenesTotal + i];
                }
                for (int i = puntoCruzaCondiciones; i < nCondicionesTotal; i++)
                {
                    nuevoCluster.genesCondiciones[nGenesTotal + i] = genesCondiciones[nGenesTotal + i];
                }
            }
            nuevoCluster.numGenes = nuevoCluster.calculaNumGenes();
            nuevoCluster.numCondiciones = nuevoCluster.calculaNumCondiciones();

            return nuevoCluster;
        }

        public int calculaNumGenes()
        {
            int nGenes = 0;
            for (int g = 0; g < nGenesTotal; g++)
            {
                if (genesCondiciones[g] == 1)
                {
                    nGenes++;
                }
            }
            return nGenes;
        }

        public int calculaNumCondiciones()
        {
            int nCondiciones = 0;
            for (int c = 0; c < nCondicionesTotal; c++)
            {
                if (genesCondiciones[nGenesTotal + c] == 1)
                {
                    nCondiciones++;
                }
            }
            return nCondiciones;
        }

        public double calculaAptitud()
        {
            return calculoFuncionVariacion();
        }

        public double calculoFuncionVariacion()  
        {
            if (numGenes < 2 || numCondiciones < 2) return 0;
            double[,] datosCluster = new double[numGenes, numCondiciones];
            int nG = 0, nC = 0;
            for (int g = 0; g < nGenesTotal; g++)
            {
                if (genesCondiciones[g] == 1)
                {
                    nC = 0;
                    for (int c = 0; c < nCondicionesTotal; c++)
                    {
                        if (genesCondiciones[nGenesTotal + c] == 1)
                        {
                            datosCluster[nG, nC] = datos[g, c];
                            nC++;
                        }
                    }
                    nG++;
                }
            }

            double[,] dMedia = new double[numGenes, numCondiciones];
            double[,] pDist = new double[numGenes, numCondiciones];

            for (int g = 0; g < numGenes; g++)
            {
                double suma_dMedia = 0;
                for (int c = 1; c < numCondiciones; c++)
                {
                    dMedia[g, c] = Math.Abs(datosCluster[g, c] - datosCluster[g, c - 1]);
                    suma_dMedia += Math.Abs(dMedia[g, c]);
                }
                for (int c = 1; c < numCondiciones; c++)
                {
                    if (suma_dMedia != 0)
                    {
                        pDist[g, c] = dMedia[g, c] / suma_dMedia;
                    }
                    else
                    {
                        pDist[g, c] = 1 / numCondiciones;
                    }
                }
            }
            double[] mProp = new double[numCondiciones];
            for (int c = 1; c < numCondiciones; c++)
            {
                mProp[c] = 0; ;
                for (int g = 0; g < numGenes; g++)
                {
                    mProp[c] += pDist[g, c];
                }
                mProp[c] /= numGenes;
            }
            double res = 0;
            for (int c = 1; c < numCondiciones; c++)
            {
                for (int g = 0; g < numGenes; g++)
                {
                    res += Math.Abs(pDist[g, c] - mProp[c]);
                }
            }
            res *= (numCondiciones - 1); 
            
            return res;
        }

    }
}
