﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Security.Cryptography;
using System.IO;

namespace Biclustering
{
    public class Poblacion
    {
        double[,] datos;
        int nGenes;
        int nCondiciones;
        public Cluster[] clusters;
        Cluster[] padres;      
        int nClusters;
        int umbralGenes;
        int umbralCondiciones;

        RNGCryptoServiceProvider rng = new RNGCryptoServiceProvider();
        byte[] dato = new byte[4];

        public Poblacion(double[,] datos, int nClusters, int umbralGenes, int umbralCondiciones)
        {
            this.datos = datos;
            nGenes = datos.GetLength(0);
            nCondiciones = datos.GetLength(1);
            this.nClusters = nClusters;
            clusters = new Cluster[nClusters];
            padres = new Cluster[nClusters];
            this.umbralGenes = umbralGenes;
            this.umbralCondiciones = umbralCondiciones;
            poblacionInicial();
        }
   
        public void poblacionInicial()
        {
            for (int nC = 0; nC < nClusters; nC++)
            {
                clusters[nC] = new Cluster(datos);
                clusters[nC].inicializa(umbralGenes,umbralCondiciones);
            }
        }

        public void calculaAptitud()  
        {
            Cluster clusterAux;
            int nClusters = clusters.Length;

            for (int c = 0; c < nClusters; c++)
            {
                clusters[c].variación = clusters[c].calculaAptitud();
                if (clusters[c].numGenes >= umbralGenes && clusters[c].numCondiciones >= umbralCondiciones)
                {
                    if (clusters[0].numGenes < umbralGenes || clusters[0].numCondiciones < umbralCondiciones || clusters[c].variación < clusters[0].variación)
                    {
                        clusterAux = clusters[0];
                        clusters[0] = clusters[c];
                        clusters[c] = clusterAux;
                    }
                }
            }
        }

        public void mutacion(int pMutacion)
        {
            for (int c = 0; c < nClusters; c++)
            {
                rng.GetNonZeroBytes(dato);
                int valor = BitConverter.ToInt32(dato, 0);
                int prob = Math.Abs(valor % 100);

                if (pMutacion > prob && c!=0)
                {
                    clusters[c].mutacion();
                }
            }
        }
      
        public void cruza(int pCruza)
        {
            barajear(padres);
            for (int c = 0; c < nClusters; c += 2)
            {
                rng.GetNonZeroBytes(dato);
                int valor = BitConverter.ToInt32(dato, 0);
                int prob = Math.Abs(valor % 100);

                if (pCruza > prob)
                {
                    if (c == 0) clusters[c] = padres[c].copia();
                    else clusters[c] = padres[c].cruza(padres[c + 1]);
                    clusters[c + 1] = padres[c + 1].cruza(padres[c]);
                }
                else
                {
                    clusters[c] = padres[c].copia();
                    clusters[c + 1] = padres[c + 1].copia();   
                }
            }
        }

        public void barajear(Cluster[] clusters)  
        {
            Cluster clusterAux;
            int c1, c2;
            for (int cont = 0; cont < nClusters; cont++)
            {
                rng.GetNonZeroBytes(dato);
                int valor = BitConverter.ToInt32(dato, 0);
                c1 = Math.Abs(valor % nClusters);

                rng.GetNonZeroBytes(dato);
                valor = BitConverter.ToInt32(dato, 0);
                c2 = Math.Abs(valor % nClusters);

                if (c1 != 0 && c2 != 0)
                {
                    clusterAux = clusters[c1];
                    clusters[c1] = clusters[c2];
                    clusters[c2] = clusterAux;
                }
            }
        }

        public void selección(int pSeleccion)   
        {
            int cP = 0;
            for (int cont = 0; cont < 2; cont++)
            {
                barajear(clusters);
                for (int c = 0; c < nClusters; c += 2)
                {
                    rng.GetNonZeroBytes(dato);
                    int valor = BitConverter.ToInt32(dato, 0);
                    int prob = Math.Abs(valor % 100);
                    if (c == 0 && cP==0)
                    {
                        padres[cP] = clusters[c].copia();
                    }
                    else if ( ( clusters[c].numGenes >= umbralGenes && clusters[c].numCondiciones >= umbralCondiciones && 
                                (clusters[c+1].numGenes < umbralGenes || clusters[c+1].numCondiciones < umbralCondiciones ) 
                              ) ||
                              ( clusters[c].numGenes >= umbralGenes && clusters[c].numCondiciones >= umbralCondiciones && 
                                clusters[c+1].numGenes >= umbralGenes && clusters[c+1].numCondiciones >= umbralCondiciones &&
                                clusters[c].variación<clusters[c+1].variación )
                            )
                    {
                        if (pSeleccion > prob)
                        {
                            padres[cP] = clusters[c].copia();
                        }
                        else
                        {
                            padres[cP] = clusters[c + 1].copia();
                        }
                    }
                    else
                    {
                        if (pSeleccion > prob)
                        {
                            padres[cP] = clusters[c + 1].copia();
                        }
                        else
                        {
                            padres[cP] = clusters[c].copia();
                        }
                    }
                    cP++;
                }
            }

        }
    }
}
